
# ICFP 2002 Competition Client Protocol Implementation
# Jp Calderone

from twisted.protocols import basic
import string, operator, astar, os

class Package:
    def __init__(self, id, weight, x, y):
        self.id, self.weight, self.x, self.y = id, weight, x, y

    def __str__(self):
        return '<Pkg #%d %du -> %d,%d>' % (self.id, self.weight, self.x, self.y)

class Robot:
    def __init__(self, id, x, y):
        self.id,self.x,self.y = id,x,y
        self.packages = []

    def __str__(self):
        return '<Robot #%s (%s, %s) with %s packages>' % (self.id, self.x, self.y, len(self.packages))

PLAIN = '.'
WATER = '~'
WALL = '#'
BASE = '@'
ROBOT = '*' # not in the spec but it helps with pathfinding

# Position -> dx, dy
COORDS = {
    'N' : (0, +1),
    'S' : (0, -1),
    'E' : (+1, 0),
    'W' : (-1, 0)
    }

OPP = {
    (0, +1) : (0, -1),
    (0, -1) : (0, +1),
    (+1, 0) : (-1, 0),
    (-1, 0) : (+1, 0)
}

ACR = {
    (0, +1) : ((+1, 0),
               (-1, 0)),
    (0, -1) : ((+1, 0),
               (-1, 0)),
    (+1, 0) : ((0, +1),
               (0, -1)),
    (-1, 0) : ((0, +1),
               (0, -1))
}
class ICFP2002ClientProtocol(basic.LineReceiver):
    delimiter = '\n'
    MAX_LENGTH = 10L**10L # because of huge packet lists
    signon = "Player"
    
    def connectionMade(self):
        self.myPackages = []            # list of Package objects
        self.knownRobots = {}           # id -> Robot object
        self.knownPackages = {}         # (x,y) -> Package object
        self.packages = {}              # id -> package
        self.mightHavePackage = []      # list of (x, y)

        self._state = 'initial'
        self.sendLine(self.signon)
        self.doneIt = 0
        self.points = 0
        self.totalPoints = 0            # Total possible points
        self.money = 0
        self.map = []

    def connectionLost(self, reason):
        print "\n\nI'm leaving the game with %d points and $%d\n%s" % (
            self.points, self.money, reason)
        if len(self.knownRobots) == 1:
            print 'I was the only robot left standing!  I win!'
        if os.environ.get('TRAINING'):
            astar.storeCache(self.map)
    
    def lineReceived(self, line):
        if self._state == 'initial':
            self.boardX, self.boardY = map(int, line.split())
            self._state = 'map'
            self.map = []
            return
        elif self._state == 'map':
            # Add wall around the whole map
            self.map.append(WALL + line.strip() + WALL)
            if len(self.map) == self.boardY:
                # Make it so the map is 0,0 based
                self.map.insert(0, WALL * (self.boardX + 2))
                self.map.append(WALL * (self.boardX + 2))
                self.extractBases()
                self._state = 'config'
                astar.loadCache(self.map)
        elif self._state == 'config':
            self.number, self.weight, self.money = map(int, line.split())
            self.initialMoney = self.money
            self._state = 'response'
            return
        elif self._state == 'response':
            if not '#' in line:
                self.died = [line]
                self._state = 'died'
                return
            robots = line[1:].split('#')
            for r in robots:
                self.parseResponsePart(r.split())
            robotnumbers = map(lambda x:int(x.split()[0]), robots)

            dead = [x for x in self.knownRobots.keys() if x not in robotnumbers]
            for d in dead:
                del self.knownRobots[d]
                
            self._state = 'packages'
            if not self.doneIt:
                self.ourFirstTime()
                self.doneIt = 1
            
            return
        elif self._state == 'packages':
            try:
                packages = map(int, line.split())
            except ValueError:
                self.died = [line]
                self._state = 'died'
                return
            while len(packages):
                self.foundPackage(*packages[:4])
                packages = packages[4:]
            if self.getPosition() in self.mightHavePackage:
                self.mightHavePackage.remove(self.getPosition())
            self.endOfTurn()
            self._state = 'response'
        elif self._state == 'died':
            self.died.append(line)
            if len(self.died) == 4:
                self.gameEnded('\n'.join(self.died))
        else:
            print 'protocol error'

    def ourFirstTime(self):
        "Called after the initial configuration has been received"
        
    def getCarryWeight(self):
        "Calculate how much we're carrying"
        return reduce(operator.add, [x.weight for x in self.myPackages], 0)

    def foundPackage(self, number, x,y, weight):
        if number not in self.packages:
            pkg = Package(number, weight, x, y)
            self.packages[number] = pkg
            self.knownPackages.setdefault(self.getPosition(),[]).append(pkg)
            self.totalPoints += weight


    def parseResponsePart(self, robot):
        number = int(robot[0])
        if len(robot) == 1:
            return # nothing to parse
        if robot[1] in COORDS.keys():
            self.robotMoved(number, robot[1])
            robot = robot[2:]
        elif robot[1] == 'X':
            self.robotAppeared(number, int(robot[2]), int(robot[4]))
            robot = robot[5:]
        elif robot[1] == 'P':
            self.robotGotPackage(number, int(robot[2]))
            robot = robot[3:]
        elif robot[1] == 'D':
            self.robotDroppedPackage(number, int(robot[2]))
            robot = robot[3:]
        if len(robot):
            self.parseResponsePart([number] + robot)


    def cellAt(self, x, y):
        if y == 0 or x == 0:
            return '#'
        if y > self.boardY or x > self.boardX:
            return '#'
        return self.map[y - 1][x - 1]


    def extractBases(self):
        # z3p > *
        self.bases = []
        for y in range(len(self.map)):
            x = -1
            while 1:
                x = self.map[y].find(BASE, x+1)
                if x == -1:
                    break
                self.bases.append((x,y))

    def moveRobot(self, bid, direction):
        self.sendLine('%d Move %s' % (int(bid), direction))
        self.money = self.money - abs(bid)


    def getPackage(self, bid, numbers):
        self.sendLine('%d Pick %s' % (int(bid), ' '.join(map(str, numbers))))
        self.money = self.money - abs(int(bid))


    def dropPackage(self, bid, numbers):
        self.sendLine('%d Drop %s' % (bid, ' '.join(map(str, numbers))))
        self.money = self.money - abs(bid)


    def robotMoved(self, robot, direction):
        dx, dy = COORDS[direction]
        self.knownRobots[robot].x += dx
        self.knownRobots[robot].y += dy

    def robotAppeared(self, robot, x, y):
        self.knownRobots[robot] = Robot(robot, x,y)
        if robot == self.number:
            self.myRobot = self.knownRobots[robot]


    def robotGotPackage(self, robot, package):
        if robot == self.number:
            self.myPackages.append(self.packages[package])
        self.knownRobots[robot].packages.append(package)

        # We may not know the package details, only the ID
        for k,v in self.knownPackages.items():
            for x in v:
                if x.id == package:
                    v.remove(x)
                    break

    def robotDroppedPackage(self, robot, package):
        r = self.knownRobots[robot]
        if robot == self.number:
            p = self.packages[package]
            self.myPackages.remove(p)
            if (p.x,p.y) == self.getPosition():
                # A winner is you!
                self.points += p.weight
        try:
            r.packages.remove(package)
        except ValueError:
            # ??? 
            pass

        # XXX Here we can only update the map if we know
        # the package has not disappeared
        p = self.packages.get(package)
        if p and p.x != r.x and p.y != r.y:
            self.knownPackages.setdefault((r.x,r.y), []).append(p)
        else:
            self.mightHavePackage.append((r.x, r.y))

    def endOfTurn(self):
        pass

    def gameEnded(self, message):
        pass

    def getPosition(self):
        "Return (x,y) of my own position"
        return self.myRobot.x, self.myRobot.y


    def getCellInDir(self, dir):
        x, y = self.getPosition()
        dx,dy = COORDS[dir]
        return self.cellAt(x + dx, y + dy)


def DumbClient(ICFP2002ClientProtocol):
    """
    Client that always goes north if possible, otherwise turns right.
    Picks up every package it finds, and drops it if it happens to arrive
    at its destination.
    """
    
    def pickBid(self):
        return max(1, self.money / 500)


    def endOfTurn(self):
        dirs = COORDS.keys()
        
        for (c, d) in zip(map(self.getCellInDir, dirs), dirs):
            if c in ('.', '@'):
                self.moveRobot(self.pickBid(), d)
                break
        print 'Failed to move!  Gnnng!'
        self.transport.loseConnection()


def showNumberedMap(m):
    "Display a map on stdout, nicely numbered"
    print "    %s" % ''.join([('%2d' % x)[0] for x in range(len(m[0]))])
    print "    %s" % (string.digits * 5)[:len(m[0])]
    for x,index in indices(m):
        print "%2d  %s" %(index, x)

def indices(l,offset=0):
    """
    Return a zipped tuple of the original passed list, and a an index, allowing
    iteration over the list and its indices simultanously. Optional offset parameter
    is base for the index.
    """
    return zip(l, range(offset,len(l)+offset))




